@@ -4,8 +4,7 @@
 -behaviour(application).
 -export([start_cli/0,
   string_to_tokens/5, tokens_to_quoted/3, 'string_to_quoted!'/5,
-  env_for_eval/1, env_for_eval/2, quoted_to_erl/2,
-  eval_forms/3, eval_quoted/3]).
+  env_for_eval/1, quoted_to_erl/2, eval_forms/3, eval_quoted/3]).
 -include("elixir.hrl").
 -define(system, 'Elixir.System').
 
@@ -55,10 +54,6 @@ start(_Type, _Args) ->
     {argv, []},
     {no_halt, false},
 
-    %% Static options
-    {bootstrap, false},
-    {identifier_tokenizer, Tokenizer},
-
     %% Compiler options
     {docs, true},
     {ignore_module_conflict, false},
@@ -71,6 +66,8 @@ start(_Type, _Args) ->
     | URIConfig
   ],
 
+  elixir_config:static(#{bootstrap => false, identifier_tokenizer => Tokenizer}),
+
   Tab = elixir_config:new(Config),
   case elixir_sup:start_link() of
     {ok, Sup} ->
@@ -111,10 +108,10 @@ preload_common_modules() ->
 parse_otp_release() ->
   %% Whenever we change this check, we should also change Makefile.
   case string:to_integer(erlang:system_info(otp_release)) of
-    {Num, _} when Num >= 21 ->
+    {Num, _} when Num >= 22 ->
       Num;
     _ ->
-      io:format(standard_error, "ERROR! Unsupported Erlang/OTP version, expected Erlang/OTP 21+~n", []),
+      io:format(standard_error, "ERROR! Unsupported Erlang/OTP version, expected Erlang/OTP 22+~n", []),
       erlang:halt(1)
   end.
 
@@ -165,10 +162,22 @@ start_cli() ->
 
 %% EVAL HOOKS
 
-env_for_eval(Opts) ->
-  env_for_eval(elixir_env:new(), Opts).
+env_for_eval(#{lexical_tracker := Pid} = Env) ->
+  NewEnv = Env#{
+    context := nil,
+    context_modules := [],
+    macro_aliases := [],
+    versioned_vars := #{}
+  },
+
+  case is_pid(Pid) andalso is_process_alive(Pid) of
+    true -> NewEnv;
+    false -> NewEnv#{lexical_tracker := nil, tracers := []}
+  end;
+%% TODO: Deprecate all options except line and file.
+env_for_eval(Opts) when is_list(Opts) ->
+  Env = elixir_env:new(),
 
-env_for_eval(Env, Opts) ->
   Line = case lists:keyfind(line, 1, Opts) of
     {line, LineOpt} when is_integer(LineOpt) -> LineOpt;
     false -> ?key(Env, line)
@@ -244,15 +253,15 @@ eval_quoted(Tree, Binding, #{line := Line} = E) ->
 
 eval_forms(Tree, Binding, Opts) when is_list(Opts) ->
   eval_forms(Tree, Binding, env_for_eval(Opts));
-eval_forms(Tree, RawBinding, OE) ->
-  {Vars, Binding} = normalize_binding(RawBinding, [], []),
-  E = elixir_env:with_vars(OE, Vars),
-  {_, S} = elixir_env:env_to_scope(E),
-  {Erl, NewE, NewS} = quoted_to_erl(Tree, E, S),
+eval_forms(Tree, RawBinding, OrigE) ->
+  {Vars, Binding} = normalize_binding(RawBinding, [], [], 0),
+  E = elixir_env:with_vars(OrigE, Vars),
+  {_, S} = elixir_env:env_to_erl(E),
+  {Erl, NewErlS, NewExS, NewE} = quoted_to_erl(Tree, E, S),
 
   case Erl of
     {atom, _, Atom} ->
-      {Atom, Binding, NewE};
+      {Atom, RawBinding, NewE};
 
     _  ->
       Exprs =
@@ -263,15 +272,15 @@ eval_forms(Tree, RawBinding, OE) ->
 
       ErlBinding = elixir_erl_var:load_binding(Binding, E, S),
       {value, Value, NewBinding} = recur_eval(Exprs, ErlBinding, NewE),
-      {Value, elixir_erl_var:dump_binding(NewBinding, NewE, NewS), NewE}
+      {Value, elixir_erl_var:dump_binding(NewBinding, NewExS, NewErlS), NewE}
   end.
 
-normalize_binding([{Key, Value} | Binding], Vars, Acc) when is_atom(Key) ->
-  normalize_binding(Binding, [{Key, nil} | Vars], [{{Key, nil}, Value} | Acc]);
-normalize_binding([{Pair, Value} | Binding], Vars, Acc) ->
-  normalize_binding(Binding, [Pair | Vars], [{Pair, Value} | Acc]);
-normalize_binding([], Vars, Acc) ->
-  {Vars, Acc}.
+normalize_binding([{Key, Value} | Binding], Vars, Acc, I) when is_atom(Key) ->
+  normalize_binding(Binding, [{{Key, nil}, I} | Vars], [{{Key, nil}, Value} | Acc], I + 1);
+normalize_binding([{Pair, Value} | Binding], Vars, Acc, I) ->
+  normalize_binding(Binding, [{Pair, I} | Vars], [{Pair, Value} | Acc], I + 1);
+normalize_binding([], Vars, Acc, _I) ->
+  {maps:from_list(Vars), Acc}.
 
 recur_eval([Expr | Exprs], Binding, Env) ->
   {value, Value, NewBinding} =
@@ -327,30 +336,31 @@ merge_stacktrace([StackItem | Stacktrace], CurrentStack) ->
 %% Converts a quoted expression to Erlang abstract format
 
 quoted_to_erl(Quoted, E) ->
-  {_, S} = elixir_env:env_to_scope(E),
+  {_, S} = elixir_env:env_to_erl(E),
   quoted_to_erl(Quoted, E, S).
 
 quoted_to_erl(Quoted, Env, Scope) ->
-  {Expanded, NewEnv} = elixir_expand:expand(Quoted, Env),
-  {Erl, NewScope} = elixir_erl_pass:translate(Expanded, Scope),
-  {Erl, NewEnv, NewScope}.
+  {Expanded, NewExS, NewEnv} = elixir_expand:expand(Quoted, elixir_env:env_to_ex(Env), Env),
+  {Erl, NewErlS} = elixir_erl_pass:translate(Expanded, erl_anno:new(?key(Env, line)), Scope),
+  {Erl, NewErlS, NewExS, NewEnv}.
 
 %% Converts a given string (charlist) into quote expression
 
 string_to_tokens(String, StartLine, StartColumn, File, Opts) when is_integer(StartLine), is_binary(File) ->
   case elixir_tokenizer:tokenize(String, StartLine, StartColumn, [{file, File} | Opts]) of
-    {ok, _Tokens} = Ok ->
-      Ok;
-    {error, {Line, Column, {ErrorPrefix, ErrorSuffix}, Token}, _Rest, _SoFar} ->
+    {ok, _Line, _Column, Warnings, Tokens} ->
+      [elixir_errors:erl_warn(L, F, M) || {L, F, M} <- lists:reverse(Warnings)],
+      {ok, Tokens};
+    {error, {Line, Column, {ErrorPrefix, ErrorSuffix}, Token}, _Rest, _Warnings, _SoFar} ->
       Location = [{line, Line}, {column, Column}],
       {error, {Location, {to_binary(ErrorPrefix), to_binary(ErrorSuffix)}, to_binary(Token)}};
-    {error, {Line, Column, Error, Token}, _Rest, _SoFar} ->
+    {error, {Line, Column, Error, Token}, _Rest, _Warnings, _SoFar} ->
       Location = [{line, Line}, {column, Column}],
       {error, {Location, to_binary(Error), to_binary(Token)}}
   end.
 
-tokens_to_quoted(Tokens, File, Opts) ->
-  handle_parsing_opts(File, Opts),
+tokens_to_quoted(Tokens, WarningFile, Opts) ->
+  handle_parsing_opts(WarningFile, Opts),
 
   try elixir_parser:parse(Tokens) of
     {ok, Forms} ->
@@ -360,7 +370,7 @@ tokens_to_quoted(Tokens, File, Opts) ->
     {error, {Line, _, [Error, Token]}} ->
       {error, {parser_location(Line), to_binary(Error), to_binary(Token)}}
   after
-    erase(elixir_parser_file),
+    erase(elixir_parser_warning_file),
     erase(elixir_parser_columns),
     erase(elixir_token_metadata),
     erase(elixir_literal_encoder)
@@ -386,17 +396,17 @@ parser_location(Meta) ->
       case tokens_to_quoted(Tokens, File, Opts) of
         {ok, Forms} ->
           Forms;
-        {error, {Line, Error, Token}} ->
-          elixir_errors:parse_error(Line, File, Error, Token)
+        {error, {Meta, Error, Token}} ->
+          elixir_errors:parse_error(Meta, File, Error, Token, {String, StartLine, StartColumn})
       end;
-    {error, {Line, Error, Token}} ->
-      elixir_errors:parse_error(Line, File, Error, Token)
+    {error, {Meta, Error, Token}} ->
+      elixir_errors:parse_error(Meta, File, Error, Token, {String, StartLine, StartColumn})
   end.
 
 to_binary(List) when is_list(List) -> elixir_utils:characters_to_binary(List);
 to_binary(Atom) when is_atom(Atom) -> atom_to_binary(Atom, utf8).
 
-handle_parsing_opts(File, Opts) ->
+handle_parsing_opts(WarningFile, Opts) ->
   LiteralEncoder =
     case lists:keyfind(literal_encoder, 1, Opts) of
       {literal_encoder, Fun} -> Fun;
@@ -404,7 +414,7 @@ handle_parsing_opts(File, Opts) ->
     end,
   TokenMetadata = lists:keyfind(token_metadata, 1, Opts) == {token_metadata, true},
   Columns = lists:keyfind(columns, 1, Opts) == {columns, true},
-  put(elixir_parser_file, File),
+  put(elixir_parser_warning_file, WarningFile),
   put(elixir_parser_columns, Columns),
   put(elixir_token_metadata, TokenMetadata),
   put(elixir_literal_encoder, LiteralEncoder).
